/* main.c
 *       Copyright (C) 2001, 2002 Timo Schulz 
 *
 * This file is part of OpenCDK.
 *
 * OpenCDK is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * OpenCDK is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with OpenCDK; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdio.h>
#include <errno.h>
#include <malloc.h>
#ifdef HAVE_PWD_H
# include <pwd.h>
#endif
#include <unistd.h>

#include "opencdk.h"
#include "main.h"

static void *(*alloc_func) (size_t n) = gcry_xmalloc;
static void *(*alloc_secure_func) (size_t n) = gcry_malloc_secure;
static void *(*realloc_func) (void *p, size_t n) = gcry_realloc;
static void *(*calloc_func) (size_t m, size_t n) = gcry_calloc;
static void (*free_func) (void *) = gcry_free;
static int malloc_hooks = 0;
static int secmem_init = 0;

static CDK_LOG_FNC log_handler = NULL;
static void *log_handler_value = NULL;
static int log_level = CDK_LOG_NONE;

static CDK_PASSPHRASE_CB passphrase_cb = NULL;


const char *
cdk_strerror (int ec)
{
  static char buf[20];

  switch (ec)
    {
    case CDK_EOF:              return "End Of File";
    case CDK_Success:          return "No error";
    case CDK_General_Error:    return "General error";
    case CDK_File_Error:       return strerror (errno);
    case CDK_Bad_Sig:          return "Bad signature";
    case CDK_Inv_Packet:       return "Invalid packet";
    case CDK_Inv_Algo:         return "Invalid algorithm";
    case CDK_Not_Implemented:  return "This is not implemented yet";
    case CDK_Gcry_Error:       return gcry_strerror (gcry_errno ());
    case CDK_Armor_Error:      return "ASCII armor error";
    case CDK_Armor_CRC_Error:  return "ASCII armored damaged (CRC error)";
    case CDK_MPI_Error:        return "Invalid or missformed MPI";
    case CDK_Inv_Value:        return "Invalid parameter or value";
    case CDK_Error_No_Key:     return "No key available";
    case CDK_Chksum_Error:     return "Check for key doesn't match";
    case CDK_Time_Conflict:    return "Time conflict";
    case CDK_Zlib_Error:       return "ZLIB error";
    case CDK_Weak_Key:         return "Weak key was detected";
    case CDK_Out_Of_Core:      return "Out of core!!";
    case CDK_Wrong_Seckey:     return "Wrong secret key";
    case CDK_Bad_MDC:          return "Manipulated MDC detected";
    case CDK_Inv_Mode:         return "Invalid mode";
    case CDK_Error_No_Keyring: return "No keyring available";
    default:                   sprintf (buf, "ec=%d", ec); return buf;
    }
  return NULL;
}


static void
out_of_core (void)
{
  fputs ("\n ** fatal error: out of memory **\n", stderr);
}


void
cdk_set_malloc_hooks (void *(*new_alloc_func) (size_t n),
		      void *(*new_alloc_secure_func) (size_t n),
		      void *(*new_realloc_func) (void *p, size_t n),
		      void *(*new_calloc_func) (size_t m, size_t n),
		      void (*new_free_func) (void *))
{
  alloc_func = new_alloc_func;
  alloc_secure_func = new_alloc_secure_func;
  realloc_func = new_realloc_func;
  calloc_func = new_calloc_func;
  free_func = new_free_func;
  malloc_hooks = 1;
}


int
cdk_malloc_hook_initialized (void)
{
  return malloc_hooks;
}


void *
cdk_malloc (size_t size)
{
  void *p = alloc_func (size);
  if (!p)
    out_of_core ();
  return p;
}


void *
cdk_calloc (size_t n, size_t m)
{
  void *p = calloc_func (n, m);
  if (!p)
    out_of_core ();
  return p;
}


static void
_secmem_init (size_t size)
{
  if (!size)
    {
      gcry_control (GCRYCTL_DROP_PRIVS);
      return;
    }
  if (secmem_init == 1)
    return;
  if (size >= 65536)
    size = 16384;
  gcry_control (GCRYCTL_INIT_SECMEM, 16384, 0);
  gcry_control (GCRYCTL_DISABLE_SECMEM_WARN);
  secmem_init = 1;
}


static void
_secmem_end (void)
{
  gcry_control (GCRYCTL_TERM_SECMEM);
  secmem_init = 0;
}


void *
cdk_salloc (size_t size)
{
  void *p;
  
  if (!secmem_init)
    {
      _secmem_init (16384);
      secmem_init = 1;
    }
  if (!secmem_init == 1)
    {
      _secmem_init (0);
      secmem_init = 2;
    }
  p = alloc_secure_func (size);
  if (!p)
    out_of_core ();
  return p;
}


void *
cdk_scalloc (size_t m)
{
  void *p = cdk_salloc (m);
  memset (p, 0, m);
  return p;
}


void *
cdk_realloc (void *ptr, size_t size)
{
  void *p = realloc_func (ptr, size);
  if (!p)
    out_of_core ();
  return p;
}


void *
cdk_strdup (const char * ptr)
{
  char * p = cdk_malloc (strlen (ptr) + 1);
  if (p)
    strcpy (p, ptr);
  return p;
}


void
cdk_free (void * ptr)
{
  if (ptr)
    free_func (ptr);
}


static void
_cdk_logv (int level, const char *fmt, va_list arg_ptr)
{

  if (log_handler)
    log_handler (log_handler_value, level, fmt, arg_ptr);
  else
    {
      switch (level)
        {
        case CDK_LOG_INFO : break;
        case CDK_LOG_DEBUG: fputs ("DBG: ", stderr); break;
        case CDK_LOG_NONE : return; 
        }
      vfprintf (stderr, fmt, arg_ptr); 
    }
}


void
cdk_set_log_handler (CDK_LOG_FNC logfnc, void *opaque)
{
  log_handler = logfnc;
  log_handler_value = opaque; 
}


void
cdk_set_log_level (int lvl)
{
  log_level = lvl;
}


int
_cdk_get_log_level (void)
{
  return log_level;
}


void
_cdk_log_info (const char *fmt, ...)
{
  va_list arg;

  if (log_level == CDK_LOG_NONE)
    return;
  va_start (arg, fmt);
  _cdk_logv (CDK_LOG_INFO, fmt, arg);
  va_end (arg);
}


void
_cdk_log_debug (const char *fmt, ...)
{
  va_list arg;
  
  if (log_level < CDK_LOG_DEBUG)
    return;
  va_start (arg, fmt);
  _cdk_logv (CDK_LOG_DEBUG, fmt, arg);
  va_end (arg);
}


void
cdk_passphrase_cb_set (CDK_PASSPHRASE_CB pass_fnc)
{
  if (pass_fnc)
    passphrase_cb = pass_fnc;
}


#ifndef HAVE_PWD_H
char *
getpass (const char * prompt)
{
  return NULL; /* fixme */
}
#endif


char *
_cdk_passphrase_get (const char * prompt)
{
  if (passphrase_cb)
    return passphrase_cb (prompt);
  return getpass (prompt);
}


void
_cdk_passphrase_free (char *pw)
{
  if (passphrase_cb)
    {
      cdk_free (pw);
      return;
    }
  /* if we use getpass, we've to free the handle with free() because
     it was *not* allocated with the gcry memory routines */
  free (pw);
}

  
int
cdk_handle_new (CDK_HD *r_ctx)
{
  CDK_HD c;
  
  if (!r_ctx)
    return CDK_Inv_Value;
  
  c = cdk_calloc (1, sizeof *c);
  if (!c)
    return CDK_Out_Of_Core;
  /* default */
  c->s2k.mode = 3;
  c->s2k.digest_algo = GCRY_MD_SHA1;
  c->s2k.cipher_algo = GCRY_CIPHER_CAST5;

  c->opt.mdc = 1;
  c->opt.compress = 1;
  c->opt.armor = 0;
  c->opt.textmode = 0;

  c->digest_algo = GCRY_MD_SHA1;
  c->cipher_algo = GCRY_CIPHER_CAST5;
  c->compress_algo = CDK_COMPRESS_ZLIB;

  *r_ctx = c;
  return 0;
}


void
cdk_handle_set_s2k (CDK_HD hd, int mode, int digest, int cipher)
{
  if (hd)
    {
      hd->s2k.mode = mode;
      hd->s2k.digest_algo = digest;
      hd->s2k.cipher_algo = cipher;
    }
}


void
cdk_handle_set_cipher (CDK_HD hd, int cipher)
{
  if (hd)
    hd->cipher_algo = cipher;
}


void
cdk_handle_set_digest (CDK_HD hd, int digest)
{
  if (hd)
    hd->digest_algo = digest;
}


void
cdk_handle_set_compress (CDK_HD hd, int algo)
{
  if (hd)
    hd->compress_algo = algo;
}


void
cdk_handle_free (CDK_HD c)
{
  if (c)
    {
      _cdk_result_verify_free (c->result.verify);
      cdk_free (c);
    }
}
