/* write-packet.c - Write OpenPGP packets
 *  Copyright (C) 2001, 2002 Timo Schulz 
 *
 * This file is part of OpenCDK.
 *
 * OpenCDK is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 *  
 * OpenCDK is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with OpenCDK; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <string.h>
#include <stdio.h>

#include "opencdk.h"
#include "main.h"


static int
stream_write (CDK_STREAM s, const void *buf, size_t count)
{
  int nwritten = cdk_stream_write (s, buf, count);
  if (nwritten == EOF)
    return CDK_File_Error;
  return 0;  
}


static int
stream_read (CDK_STREAM s, void *buf, size_t count, size_t *r_nread)
{
  int nread = cdk_stream_read (s, buf, count);
  if (nread == EOF)
    return CDK_File_Error;
  *r_nread = nread;
  return 0;
}


static int
stream_putc (CDK_STREAM s, int c)
{
  int nwritten;
  nwritten = cdk_stream_putc (s, c);
  if (nwritten == EOF)
    return CDK_File_Error;
  return 0;
}


static int
write_32 (CDK_STREAM out, u32 u)
{
  byte buf[4];

  buf[0] = u >> 24;
  buf[1] = u >> 16;
  buf[2] = u >> 8;
  buf[3] = u;
  return stream_write (out, buf, 4);
}


static int
write_16(CDK_STREAM out, u16 u)
{
  byte buf[2];

  buf[0] = u >> 8;
  buf[1] = u;
  return stream_write (out, buf, 2);
}


static size_t
calc_mpi_size (CDK_MPI mpi[4], int ncount)
{
  size_t nbytes, size;
  int i;

  size = 0;
  for (i = 0; i < ncount; i++)
    {
      nbytes = (mpi[i]->bits + 7) / 8 + 2;
      size += nbytes;
    }
  return size;
}


static int
write_mpi (CDK_STREAM out, CDK_MPI m)
{
  if (!out || !m)
    return CDK_Inv_Value;

  if (m->bits > MAX_MPI_BITS || !m->bits)
    return CDK_MPI_Error;
  if (!m->bytes)
    m->bytes = (m->bits + 7) / 8;
  return stream_write (out, m->data, m->bytes + 2);
}


static int
pkt_encode_len (CDK_STREAM out, size_t pktlen)
{
  int rc = 0;

  if (!out)
    return CDK_Inv_Value;

  if (pktlen < 192)
    rc = stream_putc (out, pktlen);
  else if (pktlen < 8384)
    {
      pktlen -= 192;
      rc = stream_putc (out, (pktlen / 256) + 192);
      if (!rc)
	rc = stream_putc (out, (pktlen % 256));
    }
  else
    {
      rc = stream_putc (out, 255);
      if (!rc)
	rc = write_32 (out, pktlen);
    }

  return rc;
}


static int
pkt_write_head (CDK_STREAM out, size_t size, int type)
{
  int rc = 0;

  if (!out)
    return CDK_Inv_Value;

  if (type < 0 || type > 63)
    return CDK_Inv_Packet;
  rc = stream_putc (out, (0xC0 | type));
  if (!rc)
    rc = pkt_encode_len (out, size);
  return rc;
}


static int
pkt_write_head2 (CDK_STREAM out, int type)
{
  int ctb;
  
  if (!out)
    return CDK_Inv_Value;
  if (type < 0 || type > 63)
    return CDK_Inv_Packet;
  ctb = 0x80 | (type << 2) | 3;
  return stream_putc (out, ctb);
}


static int
write_encrypted (CDK_STREAM out, cdkPKT_encrypted * enc)
{
  int nbytes = 0;
  int rc = 0;

  if (!out || !enc)
    return CDK_Inv_Value;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write encrypted packet %lu bytes\n", enc->len);

  nbytes = enc->len ? (enc->len + enc->extralen) : 0;
  rc = pkt_write_head (out, nbytes, CDK_PKT_ENCRYPTED);
  /* rest of the packet is ciphertext */

  return rc;
}


static int
write_encrypted_mdc (CDK_STREAM out, cdkPKT_encrypted * enc)
{
  int rc = 0;
  int nbytes = 0;

  if (!out || !enc)
    return CDK_Inv_Value;

  if (!enc->mdc_method)
    return CDK_Inv_Packet;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write encrypted mdc packet %lu bytes\n", enc->len);

  nbytes = enc->len ? (enc->len + enc->extralen + 1) : 0;
  rc = pkt_write_head (out, nbytes, CDK_PKT_ENCRYPTED_MDC);
  if (!rc)
    rc = stream_putc (out, 1);	/* version */
  /* rest of the packet is ciphertext */
  return rc;
}


static int
write_symkey_enc (CDK_STREAM out, cdkPKT_symkey_enc * ske)
{
  size_t size = 0, s2k_size = 0;
  int rc = 0;

  if (!out || !ske)
    return CDK_Inv_Value;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write symmetric key encrypted packet\n");

  if (ske->s2k.mode == 1 || ske->s2k.mode == 3)
    s2k_size = 8;
  if (ske->s2k.mode == 3)
    s2k_size++;
  size = 4 + s2k_size + ske->seskeylen;
  rc = pkt_write_head (out, size, CDK_PKT_SYMKEY_ENC);
  if (!rc)
    rc = stream_putc (out, ske->version);
  if (!rc)
    rc = stream_putc (out, ske->cipher_algo);
  if (!rc)
    rc = stream_putc (out, ske->s2k.mode);
  if (!rc)
    rc = stream_putc (out, ske->s2k.hash_algo);
  if (ske->s2k.mode == 1 || ske->s2k.mode == 3)
    {
      rc = stream_write (out, ske->s2k.salt, 8);
      if (!rc && ske->s2k.mode == 3)
	rc = stream_putc (out, ske->s2k.count);
    }
  return rc;
}


static int
write_pubkey_enc (CDK_STREAM out, cdkPKT_pubkey_enc * pke)
{
  int i, size, nenc = 0;
  int rc = 0;

  if (!out || !pke)
    return CDK_Inv_Value;

  if (!KEY_CAN_ENCRYPT (pke->pubkey_algo))
    return CDK_Inv_Algo;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write public key encrypted packet\n");

  nenc = cdk_pk_get_nenc (pke->pubkey_algo);
  size = 10 + calc_mpi_size (pke->mpi, nenc);
  rc = pkt_write_head (out, size, CDK_PKT_PUBKEY_ENC);
  if (!rc)
    rc = stream_putc (out, 3);
  if (!rc)
    rc = write_32 (out, pke->keyid[0]);
  if (!rc)
    rc = write_32 (out, pke->keyid[1]);
  if (!rc)
    rc = stream_putc (out, pke->pubkey_algo);
  for (i = 0; i < nenc; i++)
    {
      rc = write_mpi (out, pke->mpi[i]);
      if (rc)
	break;
    }
  return rc;
}


static int
write_mdc (CDK_STREAM out, cdkPKT_mdc * mdc)
{
  int rc = 0;

  if (!out || !mdc)
    return CDK_Inv_Value;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write_mdc\n");

  /* This packet requires a fixed header encoding */
  rc = stream_putc (out, 0xd3);	/* packet ID and 1 byte length */
  if (!rc)
    rc = stream_putc (out, 0x14);
  if (!rc)
    rc = stream_write (out, mdc->hash, sizeof mdc->hash);
  return rc;
}


static int
subpkt_calc_size (CDK_SUBPKT s)
{
  size_t nbytes;
  
  _cdk_subpkt_get_array (s, 1, &nbytes);
  return nbytes;
}


static int
write_signature (CDK_STREAM out, cdkPKT_signature * sig)
{
  byte *buf;
  size_t nbytes;
  int i, size, nsig = 0;
  int rc = 0;

  if (!out || !sig)
    return CDK_Inv_Value;

  if (!KEY_CAN_SIGN (sig->pubkey_algo))
    return CDK_Inv_Algo;
  else if (sig->version < 3 || sig->version > 4)
    return CDK_Inv_Packet;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write signature packet\n");

  nsig = cdk_pk_get_nsig (sig->pubkey_algo);
  if (sig->version < 4)
    {
      size = 19 + calc_mpi_size (sig->mpi, nsig);
      rc = pkt_write_head (out, size, CDK_PKT_SIGNATURE);
      if (!rc)
	rc = stream_putc (out, 3);
      if (!rc)
	rc = stream_putc (out, 5);
      if (rc)
	rc = stream_putc (out, sig->sig_class);
      if (!rc)
	rc = write_32 (out, sig->timestamp);
      if (!rc)
	rc = write_32 (out, sig->keyid[0]);
      if (!rc)
	rc = write_32 (out, sig->keyid[1]);
      if (!rc)
	rc = stream_putc (out, sig->pubkey_algo);
      if (!rc)
	rc = stream_putc (out, sig->digest_algo);
      if (!rc)
	rc = stream_putc (out, sig->digest_start[0]);
      if (!rc)
	rc = stream_putc (out, sig->digest_start[1]);
    }
  else
    {
      size = 10 + subpkt_calc_size (sig->hashed)
	+ subpkt_calc_size (sig->unhashed)
	+ calc_mpi_size (sig->mpi, nsig);
      rc = pkt_write_head (out, size, CDK_PKT_SIGNATURE);
      if (!rc)
	rc = stream_putc (out, 4);
      if (!rc)
	rc = stream_putc (out, sig->sig_class);
      if (!rc)
	rc = stream_putc (out, sig->pubkey_algo);
      if (!rc)
	rc = stream_putc (out, sig->digest_algo);
      if (!rc)
	rc = write_16 (out, sig->hashed_size);
      if (!rc)
        {
          buf = _cdk_subpkt_get_array (sig->hashed, 0, &nbytes);
          if (!buf)
            return CDK_Out_Of_Core;
          rc = stream_write (out, buf, nbytes);
          cdk_free (buf);
        }
      if (!rc)
	rc = write_16 (out, sig->unhashed_size);
      if (!rc)
        {
          buf = _cdk_subpkt_get_array (sig->unhashed, 0, &nbytes);
          if (!buf)
            return CDK_Out_Of_Core;
          rc = stream_write (out, buf, nbytes);
          cdk_free (buf);
        }
      if (!rc)
	rc = stream_putc (out, sig->digest_start[0]);
      if (!rc)
	rc = stream_putc (out, sig->digest_start[1]);
    }

  for (i = 0; i < nsig; i++)
    {
      rc = write_mpi (out, sig->mpi[i]);
      if (rc)
	break;
    }

  return rc;
}


static int
write_public_key (CDK_STREAM out, cdkPKT_public_key * pk)
{
  int rc = 0;
  int i, size = 6;
  size_t npkey = 0;

  if (!out || !pk)
    return CDK_Inv_Value;

  if (pk->version < 3 || pk->version > 4)
    return CDK_Inv_Packet;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write public key packet\n");

  npkey = cdk_pk_get_npkey (pk->pubkey_algo);
  if (pk->version < 4)
    size += 2;			/* expire date */
  size += calc_mpi_size (pk->mpi, npkey);
  rc = pkt_write_head (out, size, CDK_PKT_PUBLIC_KEY);
  if (!rc)
    rc = stream_putc (out, pk->version);
  if (!rc)
    rc = write_32 (out, pk->timestamp);
  if (!rc && pk->version < 4)
    rc = write_16 (out, pk->expiredate);
  if (!rc)
    rc = stream_putc (out, pk->pubkey_algo);
  for (i = 0; i < npkey; i++)
    {
      rc = write_mpi (out, pk->mpi[i]);
      if (rc)
	break;
    }
  return rc;
}


static int
write_secret_key (CDK_STREAM out, cdkPKT_secret_key * sk)
{
  cdkPKT_public_key *pk = NULL;
  size_t size = 0, npkey, nskey;
  int i = 0;
  int rc = 0;

  if (!out || !sk || !sk->pk)
    return CDK_Inv_Value;

  pk = sk->pk;
  if (pk->version < 3 || pk->version > 4)
    return CDK_Inv_Packet;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write secret key packet\n");

  npkey = cdk_pk_get_npkey (pk->pubkey_algo);
  nskey = cdk_pk_get_nskey (pk->pubkey_algo);

  rc = pkt_write_head (out, size, CDK_PKT_SECRET_KEY);
  if (!rc)
    rc = stream_putc (out, pk->version);
  if (!rc)
    rc = write_32 (out, pk->timestamp);
  if (!rc && pk->version < 4)
    {
      u16 ndays = 0;
      if (pk->expiredate)
	ndays = (u16) ((pk->expiredate - pk->timestamp) / 86400L);
      rc = write_16 (out, ndays);
    }
  if (!rc)
    rc = stream_putc (out, pk->pubkey_algo);
  for (i = 0; i < npkey && !rc; i++)
    rc = write_mpi (out, pk->mpi[i]);
  if (sk->is_protected == 0)
    rc = stream_putc (out, 0);
  else
    {
      if (is_RSA (pk->pubkey_algo) && pk->version < 4)
	stream_putc (out, sk->protect.algo);
      else
	{
	  rc = stream_putc (out, 0xff);
	  if (!rc)
	    rc = stream_putc (out, sk->protect.s2k.mode);
	  if (!rc)
	    rc = stream_putc (out, sk->protect.s2k.hash_algo);
	  if (!rc && (sk->protect.s2k.mode == 1 || sk->protect.s2k.mode == 3))
	    {
	      rc = stream_write (out, sk->protect.s2k.salt, 8);
	      if (!rc && sk->protect.s2k.mode == 3)
		rc = stream_putc (out, sk->protect.s2k.count);
	    }
	}
      rc = stream_write (out, sk->protect.iv, sk->protect.ivlen);
    }
  if (!rc && sk->is_protected && pk->version > 4)
    {
      if (sk->encdata && sk->enclen)
	rc = stream_write (out, sk->encdata, sk->enclen);
    }
  else
    {
      for (i = 0; i < nskey; i++)
	{
	  rc = write_mpi (out, sk->mpi[i]);
	  if (rc)
	    break;
	}
      if (!rc)
	rc = write_16 (out, sk->csum);
    }

  return rc;
}


static int
write_compressed (CDK_STREAM out, cdkPKT_compressed * cd)
{
  int rc;

  if (!out || !cd)
    return CDK_Inv_Value;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write compressed packet\n");

  rc = pkt_write_head2 (out, CDK_PKT_COMPRESSED);
  if (!rc)
    rc = stream_putc (out, cd->algorithm);
  return rc;
}


static int
write_plaintext (CDK_STREAM out, cdkPKT_plaintext * pt)
{
  byte buf[4096];
  size_t size;
  int rc;

  if (!out || !pt)
    return CDK_Inv_Value;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write plaintext packet\n");

  size = 6 + pt->namelen + pt->len;
  rc = pkt_write_head (out, size, CDK_PKT_PLAINTEXT);
  if (!rc)
    rc = stream_putc (out, pt->mode);
  if (!rc)
    rc = stream_putc (out, pt->namelen);
  if (!rc && pt->namelen)
    rc = stream_write (out, pt->name, pt->namelen);
  if (!rc)
    rc = write_32 (out, pt->timestamp);
  if (!rc)
    {
      while (pt->len && !cdk_stream_eof (pt->buf))
        {
          rc = stream_read (pt->buf, buf, sizeof buf-1, &size);
          if (!rc)
            rc = stream_write (out, buf, size);
          if (rc)
            break;
        }
    }
  return rc;
}

  
static int
write_onepass_sig (CDK_STREAM out, cdkPKT_onepass_sig * sig)
{
  size_t size = 0;
  int rc = 0;

  if (!out || !sig)
    return CDK_Inv_Value;

  if (DEBUG_PKT)
    _cdk_log_debug ("** write one pass signature packet\n");

  size = 15;
  rc = pkt_write_head (out, size, CDK_PKT_ONEPASS_SIG);
  if (!rc && sig->version != 3)
    return CDK_Inv_Packet;
  if (!rc)
    rc = stream_putc (out, sig->version);
  if (!rc)
    rc = stream_putc (out, sig->sig_class);
  if (!rc)
    rc = stream_putc (out, sig->digest_algo);
  if (!rc)
    rc = stream_putc (out, sig->pubkey_algo);
  if (!rc)
    rc = write_32 (out, sig->keyid[0]);
  if (!rc)
    rc = write_32 (out, sig->keyid[1]);
  if (!rc)
    rc = stream_putc (out, sig->last);
  return rc;
}


static int
write_user_id (CDK_STREAM out, cdkPKT_user_id * id)
{
  int rc = 0;

  if (!out || !id)
    return CDK_Inv_Value;

  if (id->attrib_img)
    ;/* todo */
  else
    {
      rc = pkt_write_head (out, id->len, CDK_PKT_USER_ID);
      if (!rc)
	rc = stream_write (out, id->name, id->len);
    }
  return rc;
}


int
cdk_pkt_build (CDK_STREAM out, CDK_PACKET * pkt)
{
  int rc = 0;

  if (!out || !pkt)
    return CDK_Inv_Value;

  switch (pkt->pkttype)
    {
    case CDK_PKT_PLAINTEXT:
      rc = write_plaintext (out, pkt->pkt.plaintext);
      break;
    case CDK_PKT_ONEPASS_SIG:
      rc = write_onepass_sig (out, pkt->pkt.onepass_sig);
      break;
    case CDK_PKT_MDC:
      rc = write_mdc (out, pkt->pkt.mdc);
      break;
    case CDK_PKT_SYMKEY_ENC:
      rc = write_symkey_enc (out, pkt->pkt.symkey_enc);
      break;
    case CDK_PKT_ENCRYPTED:
      rc = write_encrypted (out, pkt->pkt.encrypted);
      break;
    case CDK_PKT_ENCRYPTED_MDC:
      rc = write_encrypted_mdc (out, pkt->pkt.encrypted);
      break;
    case CDK_PKT_PUBKEY_ENC:
      rc = write_pubkey_enc (out, pkt->pkt.pubkey_enc);
      break;
    case CDK_PKT_SIGNATURE:
      rc = write_signature (out, pkt->pkt.signature);
      break;
    case CDK_PKT_PUBLIC_KEY:
      rc = write_public_key (out, pkt->pkt.public_key);
      break;
    case CDK_PKT_PUBLIC_SUBKEY:
      rc = write_public_key (out, pkt->pkt.public_key);
      break;
    case CDK_PKT_COMPRESSED:
      rc = write_compressed (out, pkt->pkt.compressed);
      break;
    case CDK_PKT_SECRET_KEY:
      rc = write_secret_key (out, pkt->pkt.secret_key);
      break;
    case CDK_PKT_SECRET_SUBKEY:
      rc = write_secret_key (out, pkt->pkt.secret_key);
      break;
    case CDK_PKT_USER_ID:
      rc = write_user_id (out, pkt->pkt.user_id);
      break;
    default:
      rc = CDK_Inv_Packet;
      break;
    }
  return rc;
}
